<?php
/**
 * @brief Classe ECATP Easy Communication Appliance Techonology Protocol
 * @detail ECATP gère le transport et la communication entre le client et le serveur via le protocol ECATP.
 * 		   les commandes de base sont
 * 				-REGISTER
 * 				-EXEC
 * 				-RESETlibs
 * 				-ALIVE
 */
include_once LIB_ESIA_EXCEPTION;
include_once LIB_WEBSOCKET;
include_once LIB_TASK;

class ECATP extends Threaded
{
	const REGISTER_SUCCESS = 0;
	const REGISTER_FAILED = 1;
	/** @brief active ou désactive le debug mode */ 
	private $debug;
	/** @brief un objet web socket */
	private $wssocket;
	/** @brief la clé par défaut de connexion */
	private $default_key = "test";
	/** @brief la méthode de callback appellée lorsqu'on reçoit une commande EXEC de ECATP */
	private $callback_exec;
	/** @brief la méthode de callback appellée lorsqu'on reçoit une commande EXEC-RESPONSE de ECATP */
	private $callback_exec_response;
	/** @brief la méthode de callback appellée lorsqu'on reçoit une commande REGISTER de ECATP */
	private $callback_register;
	/** @brief indique si la gesa est bien enregistrée	 */
	private $register;
	/** @brief Numéros de série de la GESA nécessaire pour le serveur */
	private $sn;
	/**
	 * @brief Constructeur par défaut
	 * @detail Le paramètre du constructeur est un tableau associatif pouvvant contenir différente clé.
	 * <ul>
			<li>La clé HOST et PORT sont spécifiée alors ECAPT initialisera la websocket de connextion (ajoutera le protocol ws://)</li>
			<li>La clé SOCKET est spécifié alors ECATP sera initialisé directement avec la socketStream passé en paramètre</li>
			<li>La clé URL,elle initialisera la websocket de connextion</li>
			</ul>
	 * @param inconnu $array un tableau pouvant contenir différent champs (voir le détail)
	 * @param array $ssl_options (optionnel) spécifie le contexte ssl voir SocketStream::setContextOption()
	 * @throws ECATPException 
		<table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP</td>
				<td>1</td>
				<td>BAD PARAMETER </td>	
				<td>le tableau $array n'est pas correctement inititalisé (vérifie tes clé)</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>200</td>
				<td> multiple </td>	
				<td>Si on souhaite une Socket SSL et que l'accept à donné une StreamSocket non SSL.</td></tr> 
			</tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>201</td>
				<td>multiple</td>	
				<td>Si le header HTTP n'est pas un Header compatible WebSocket.</td></tr> 
			</tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>202</td>
				<td>multiple</td>	
				<td>Si la réponse du serveur ECATP n'est pas une réponse WebSocket correcte</td></tr>
			</tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>203</td>
				<td>multiple</td>	
				<td>Si la clé envoyé par le client n'est pas acceptée.</td></tr> 
			</tr> 
		</table>
	 */
	public function __construct($array,$ssl_options=NULL)
	{
		$this->debug=false;
		$this->register=false;
		$this->sn=null;
		try 
		{
			if(array_key_exists('HOST', $array) && array_key_exists('PORT', $array))
			{
				$master=new WSClientInitiator($this->default_key,array("ecatp"),$ssl_options);
				$this->wssocket=$master->initiateConnection('wss://'.$array['HOST'].':'.$array['PORT']);
			}
			elseif(array_key_exists('SOCKET', $array))
			{
				$this->wssocket = $array['SOCKET'];
			}
			elseif(array_key_exists('URL', $array))
			{
				$master=new WSClientInitiator($this->default_key,array("ecatp"),$ssl_options);
				$this->wssocket=$master->initiateConnection($array['URL']);
			}
			else
			{
				throw new ECATPException("ECATP", "BAD PARAMETER", "1");
			}
		}
		catch(WebSocketSSLException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "200");
		}
		catch(BadHTTPResponseException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "201");
		}
		catch(BadHeaderException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "202");
		}		
		catch(BadKeysException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "203");
		}
		$this->wssocket->setDefaultBlockingMode(TRUE);
	//	echo "IP LOCAL:".$this->wssocket->getLocalAddr()." ".$this->wssocket->getPeerAddr()."\n";
	}
	/**
	 * Destructeur de la classe ECATP
	 */
	public function __destruct (  )
	{

	}
	/**
	 * @brief active/désactive le mode debug
	 * @param boolean $bool si TRUE mode debug activé
	 * 						si FALSE mode debug coupé
	 */
	public function DebugMode($bool)
	{
		$this->debug = $bool;
	}
	/**
	 * @brief Spécifie fonction de Callback
	 * @detail La fonction de callBack sera appelée dès qu'une commande de type EXEC d'ECATP sera reçue.
	 * 			Les paramètres de la fonction de callback sera obligatoirement dans l'ordre:
	 * 					$taskid			=> id de la tâche
	 * 					$name 			=> nom de la tâche
	 * 					$parameter		=> paramètre de la tâche
	 * 					$priority		=> priorité de la tâche
	 * @param string $callbackfct le nom de la function de callback
	 */
	public function setExecCallBackFunction($callbackfct)
	{
		$this->callback_exec = $callbackfct;
	}
	/**
	 * @brief Spécifie fonction de Callback
	 * @detail La fonction de callBack sera appelée dès qu'une commande de type EXEC-RESPONSE d'ECATP sera reçue.
	 * 			Les paramètres de la fonction de callback sera obligatoirement un objet ECATP
	 * 					$taskid			=> id de la tâche
	 * 					$status			=> le status de l'execution (voir Task::getExecutionStatus() ) 
	 * 					$output			=> la sortie standert de la tache
	 * 					$return			=> le code de retour de la tâche
	 * @param string $callbackfct le nom de la function de callback
	 */
	public function setExecResponseCallBackFunction($callbackfct)
	{
		$this->callback_exec_response = $callbackfct;
	}
	/**
	 * @brief Spécifie fonction de Callback
	 * @detail La fonction de callBack sera appelée dès qu'une commande de type EXEC d'ECATP sera reçue.
	 * 			Les paramètres de la fonction de callback sera obligatoirement un objet ECATP et le deuxième un objet Task
	 * @param string $callbackfct le nom de la function de callback
	 */
	public function setRegisterCallBackFunction($callbackfct)
	{
		$this->callback_register = $callbackfct;
	}
	/**
	 * @brief envoit un ALIVE au serveur
	 * @detail envoi sur la websocket une requête de type ALIVE
	 * @throws ECATPException 
	 * <table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP-WebSocket</td>
				<td>210</td>
				<td>web socket fermé</td>	
				<td>Si la SocketStream est fermée.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>211</td>
				<td>erreur de fwrite  </td>	
				<td>Si l'écriture des données à échouer.</td></tr> 
			<td>ECATP-WebSocket</td>
				<td>212</td>
				<td>erreur de fread  </td>	
				<td>Si la lecture des données à échouer.</td></tr> 
			</tr> 
	</table>
	 */
	public function sendAlive()
	{
		try 
		{
			$this->wssocket->ping(null, 'ALIVE');
		}
		catch(SocketClosedException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "210");
		}
		catch(SocketWriteException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "211");
		}
		catch(SocketReadException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "212");
		}
	}
	/**
	 * @brief envoit une demande d'execution
	 * @detail Envoit via la webSocket une demande d'exécution d'un plugin.
	 * @param object/integer $task soit un objet Task soit le TASKID mais les autre valeurs doivent être spécifiée 
	 * @param string $name (optionnel) si $task est un objet "Task". $name = le nom de la tâche
	 * @param string $parameter (optionnel) les paramètres de la tâche
	 * @param integer $priority (default=7) la priorité de la tache
	 * @throws ECATPException 
	 * <table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP</td>
				<td>2</td>
				<td>Task name is empty</td>	
				<td>Le nom de la tâche n'a pas été spécifié, l'exception n'est levée qui si $task n'est pas un objet de type Task)</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>210</td>
				<td>web socket fermé</td>	
				<td>Si la SocketStream est fermée.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>211</td>
				<td>erreur de fwrite  </td>	
				<td>Si l'écriture des données à échouer.</td></tr> 
		</table>
	 */
	public function sendExec($task,$name=null,$parameter=null,$priority=7 )
	{
		$array=array();
		if(is_object($task))
		{
			if(@get_class($task)=="Task" )
			{
				$array['TASKID'] = $task->getTaskID();
				$array['NAME'] = $task->getName();
				$array['PARAMETER'] = $task->getParameter();
				$array['PRIORITY'] = $task->getPriority();
			}
		}
		else
		{
			if(empty($name))
			{
				throw new ECATPException('ECATP', 'Task name is empty', 2);
			}
			$array['TASKID'] = $task;
			$array['NAME'] =$name;
			$array['PARAMETER'] = $parameter;
			$array['PRIORITY'] = $priority;
		}
		$ECATP['COMMAND']="EXEC";
		$ECATP['TASK']=$array;
		//echo "SEND_REQUEST\n";
		$this->sendRequest($ECATP);
	}
	/**
	 * @brief envoit la réponse à une demande d'execution
	 * @detail Envoit via la webSocket l'état de l'execution d'un plugin demandé
	 * @param object/integer $task soit un objet Task soit le TASKID mais les autre valeurs doivent être spécifiée 
	 * @param string $output (optionnel) la sortie standart du plugin 
	 * @param integer $return (optionnel)le code de retour du plugin (de 0 à 4) si $return est null alors c'est que l'execution ne sait pas bien passée (voir $status)
	 * @param interger $status (optionnel) le status de l'execution voir les constante de la classe Task. Vous pouvez avoir un message plus clair avec Task::ExecutionMessage()
	 * @throws ECATPException 
	 * <table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP-WebSocket</td>
				<td>210</td>
				<td>web socket fermé</td>	
				<td>Si la SocketStream est fermée.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>211</td>
				<td>erreur de fwrite  </td>	
				<td>Si l'écriture des données à échouer.</td></tr> 
		</table>
	 */
	public function sendExecResponse($task,$output=null,$return=null,$status=null )
	{
		$sta = $status;
		$array=array();
		if(is_object($task))
		{
			if(@get_class($task)=="Task" )
			{ 
				$array['TASKID'] = $task->getTaskID() ;
				$array['OUTPUT'] = $task->getOutput();
				$array['RETURN'] = $task->getReturnCode();
				$sta = $task->getExecutionStatus();
			}
		}
		else
		{
			$array['TASKID'] = $task;
			$array['OUTPUT'] =$output;
			$array['RETURN'] = $return;
		}
		$arr_tmp['COMMAND']="EXEC-RESPONSE";
		$arr_tmp['STATUS']=$sta;
		$arr_tmp['TASK']=$array;
		$this->sendRequest($arr_tmp);

	}
	/**
	 * @brief envoit une requête d'enregistrement au serveur
	 * @param string $sn le numéros de série de la gesa
	 * @throws ECATPException 
	 * <table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP</td>
				<td>4</td>
				<td>Register reject by server</td>	
				<td>le numéros de série ne correspond pas et le serveur à rejeté la demande.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>210</td>
				<td>web socket fermé</td>	
				<td>Si la SocketStream est fermée.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>211</td>
				<td>erreur de fwrite  </td>	
				<td>Si l'écriture des données à échouer.</td></tr> 
		</table>
	 */
	public function sendRegister($sn)
	{
		$arr_tmp['COMMAND']="REGISTER";
		echo "IP:".$this->wssocket->getLocalAddr()."\n";
		$arr_tmp['IP']=$this->wssocket->getLocalAddr();
		$arr_tmp['SN']=$sn;
		//on envoit la request
		$this->sendRequest($arr_tmp);
		//en attente de la réponse
		$return = $this->wssocket->receive(array($this,'WSSocket_Callbackreceive'));
		//pour la suite de l'authentification il faut aller voir la fonction de callback à la qui commence a cette ligne:
			//if($arr['COMMAND'] == 'REGISTER-RESPONSE')
	}
	/**
	 * 
	 * @param unknown $state
	 * @throws ECATPException 
	 * <table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP</td>
				<td>3</td>
				<td>Bad register status</td>	
				<td>le status de l'enregistrement n'est pas égal à ECATP::REGISTER_SUCCESS ou ECATP::REGISTER_FAILED</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>210</td>
				<td>web socket fermé</td>	
				<td>Si la SocketStream est fermée.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>211</td>
				<td>erreur de fwrite  </td>	
				<td>Si l'écriture des données à échouer.</td></tr> 
		</table>
	 */
	public function sendRegisterResponse($state)
	{
		if($state != ECATP::REGISTER_SUCCESS && $state != ECATP::REGISTER_FAILED)
			throw new ECATPException("ECATP","Bad register status","3");
		$arr_tmp['COMMAND']="REGISTER-RESPONSE";
		$arr_tmp['STATUS']="$state";
		$this->sendRequest($arr_tmp);
		//si l'authentification a reussie on passe en mode non bloquant
		if($state == ECATP::REGISTER_SUCCESS)
		{
			$this->register = true;
			$this->wssocket->setDefaultBlockingMode(FALSE);
		}
	}
	/**
	 * @brief envoit les données au serveur 
	 * @param array $arr_tmp un tableau associatif contenant la requete du protocol
	 * @throws ECATPException 
	 * <table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP-WebSocket</td>
				<td>210</td>
				<td>web socket fermé</td>	
				<td>Si la SocketStream est fermée.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>211</td>
				<td>erreur de fwrite  </td>	
				<td>Si l'écriture des données à échouer.</td></tr> 
		</table>
	 */
	private function sendRequest($arr_tmp)
	{
		try
		{
			$json = json_encode($arr_tmp);
			if($this->debug == true)
			{
				echo "ECAPT-SEND: $json\n";
			}
			$this->wssocket->send($json);
			$json =null;
		}
		catch(SocketClosedException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(),"210");
		}
		catch(SocketWriteException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(),"211");
		}
	}
	/**
	 * @brief méthode de callback 
	 * @detail méthode executée lors que la web socket reçoit une donnée ou est fermée.
	 * @param SocketStream $socket
	 * @param string $opcode l'opcode de la websocket
	 * @param string $data les données reçues
	 * @param integer $close_state status de fermeture voir WSSocket::close()
	 */
	public function WSSocket_Callbackreceive($socket,$opcode,$data,$close_state)
	{
		if($this->debug == true)
		{
			echo "ECAPT: $opcode $data $close_state\n";
		}
		//si les données sont bien du text on va décoder le json
		if($opcode=='text')
		{
			//on decode le json
			$arr = json_decode($data,TRUE);
			//on a une clé de commande
			if(array_key_exists('COMMAND', $arr))
			{
				//si la clé COMMAND est égal à EXEC on test pour savoir si on a un tableau TASK
				if($arr['COMMAND'] == 'EXEC' && array_key_exists('TASK', $arr))
				{
					//si on a spécifé une méthode de callback
					if(isset($this->callback_exec))
					{
						//on instancie un objet TASK et appelle la fonction de callback
						call_user_func($this->callback_exec,$arr['TASK']['TASKID'],$arr['TASK']['NAME'],$arr['TASK']['PARAMETER'],$arr['TASK']['PRIORITY']);
					}
				}
				//si la clé COMMAND est égal à EXEC on test pour savoir si on a un tableau TASK
				if($arr['COMMAND'] == 'EXEC-RESPONSE' && array_key_exists('TASK', $arr))
				{
					//si on a spécifé une méthode de callback
					if(isset($this->callback_exec_response))
					{
						if(is_string($this->callback_exec_response))
						{
							call_user_func($this->callback_exec_response,$arr['TASK']['TASKID'],$arr['STATUS'],$arr['TASK']['OUTPUT'],$arr['TASK']['RETURN']);
							//on instancie un objet TASK et appelle la fonction de callback
						}
						else
						{
							call_user_func(array($this->callback_exec_response[0],$this->callback_exec_response[1]),$arr['TASK']['TASKID'],$arr['STATUS'],$arr['TASK']['OUTPUT'],$arr['TASK']['RETURN']);
						}
					}
				}
				elseif($arr['COMMAND'] == 'REGISTER' && array_key_exists('SN', $arr))
				{
					$this->sn = $arr['SN'];
					//si on a spécifé une méthode de callback
					if(isset($this->callback_register))
					{
						//on appelle la fonction de callback
						call_user_func($this->callback_register,$this,$arr['SN'],$arr['IP']);
					}
				}
				elseif($arr['COMMAND'] == 'REGISTER-RESPONSE')
				{
					//echo "STATUS:".$arr['STATUS']."|";
					if($arr['STATUS'] != ECATP::REGISTER_SUCCESS)
						throw new ECATPException("ECATP","Register Failed by server","4");
						//authentification reussie on passe en mode non bloquant
					$this->wssocket->setDefaultBlockingMode(FALSE);
					$this->register = true;
					//return $arr['STATUS'];
				}
			}
			$arr=null;
		}	
		elseif($opcode=='close')
		{
			echo "CLOSE: $close_state\n";
		}
		elseif ($opcode=='ping')
		{
			echo "PING\n";
		}
		elseif ($opcode=='pong')
		{
			echo "PONG\n";
		}
		if( $close_state != null ) exit(0);
	}
	/**
	 * @brief indique si la GESA est enregistrée
	 * @return boolean VRAI si la GESA est enregistrée sinon false
	 */
	public function isRegister()
	{
		return $this->register;
	}
	/**
	 * @brief retourne le numéros de série de la GESA
	 * @detail cette fonction ne utilisable que sur la partie serveur d'Ecapt, 
	 * 		   la variable est affectée lors d'un "REGISTER" d'ECATP dans la fonction ECATP::WSSocket_Callbackrecive()
	 * @return integer numéros de série de la gesa si envoyé sinon null
	 */
	public function getGesaSerialNumber()
	{
		return $this->sn;
	}
	/**
	 * @detail regarde s'il y a une donnée à lire sur la socket
	 * @throws ECATPException
	 * <table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP-WebSocket</td>
				<td>210</td>
				<td>web socket fermé</td>	
				<td>Si la SocketStream est fermée.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>212</td>
				<td>erreur de fread  </td>	
				<td>Si la lecture des données à échouer.</td></tr> 
		</table>
	 */
	public function receive() 
	{
		try
		{
			$this->wssocket->receive(array($this,'WSSocket_Callbackreceive'));
		}
		catch(SocketClosedException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "210");
		}
		catch(SocketReadException $e)
		{
			throw new ECATPException("ECATP-WebSocket",$e->getMessage(), "212");
		}
	}
	/**
	 * @brieg ferme la socket
	 * @detail ferme propremet la socket si elle est connectée
	 * 
	 */
	public function close()
	{
		if ($this->wssocket->isConnected())
		{
			try 
			{
				$this->wssocket->close();
			}
			catch(SocketClosedException $e) { }
			catch(SocketReadException $e) { }
			catch(SocketWriteException $e) { }
		}
	}
	public function forceClose()
	{
		//echo "CLOSE\n";
		//$this->wssocket->destroy();
	}
	public function run() {}
	
}
/**
 * @brief classe de gestion d'Exception d'ECATP
 * @author saloc
 * @detail cette classe hérite de EsiaException
 * voici les différentes valeurs/ code
 * <table>
			<tr>
				<th>Source</th>
				<th>code</th>
				<th>Message</th>
				<th>Description :</th>
			</tr>
			<tr>
				<td>ECATP</td>
				<td>1</td>
				<td>BAD PARAMETER </td>	
				<td>le tableau $array n'est pas correctement inititalisé (vérifie tes clé)</td></tr> 
			<tr>
				<td>ECATP</td>
				<td>2</td>
				<td>Task name is empty</td>	
				<td>Le nom de la tâche n'a pas été spécifié, l'exception n'est levée qui si $task n'est pas un objet de type Task)</td></tr> 
			<tr>
				<td>ECATP</td>
				<td>3</td>
				<td>Bad register status</td>	
				<td>le status de l'enregistrement n'est pas égal à ECATP::REGISTER_SUCCESS ou ECATP::REGISTER_FAILED</td></tr> 
			<tr>
				<td>ECATP</td>
				<td>4</td>
				<td>Register reject by server</td>	
				<td>le numéros de série ne correspond pas et le serveur à rejeté la demande.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>200</td>
				<td> multiple </td>	
				<td>Si on souhaite une Socket SSL et que l'accept à donné une StreamSocket non SSL.</td></tr> 
			</tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>201</td>
				<td>multiple</td>	
				<td>Si le header HTTP n'est pas un Header compatible WebSocket.</td></tr> 
			</tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>202</td>
				<td>multiple</td>	
				<td>Si la réponse du serveur n'est pas une réponse WebSocket correcte</td></tr> 
			</tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>203</td>
				<td>multiple</td>	
				<td>Si la clé envoyé par le client n'est pas acceptée.</td></tr> 
			</tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>210</td>
				<td>web socket fermé</td>	
				<td>Si la SocketStream est fermée.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>211</td>
				<td>erreur de fwrite  </td>	
				<td>Si l'écriture des données à échouer.</td></tr> 
			<tr>
				<td>ECATP-WebSocket</td>
				<td>212</td>
				<td>erreur de fread  </td>	
				<td>Si la lecture des données à échouer.</td></tr> 
		</table>
 */
class ECATPException extends EsiaException {
}
?>
