<?php
/**
	@file
	@brief Ce fichier contient la classe EventManager
	@details Contient les classes permettant de gérer les événemets de l'application
	@author Julien George
*/

/**
 * @brief Gére les événements de l'application
 * @details Initialise l'eventhandler de php avec lui-même et envoit des notifications à des objects listeners.
 * @author Julien George
 * @todo Créer une interface que les listeners devront implémenter
 * @todo Vérifier la gestion des evénements qui doivent ou pas être envoyer au listener.
 */
	class EventManager{
			/** @brief Contient, sous forme de timestamp avec microsecondes, l'heure à laquelle l'objet fut construit. */
			private $chrono=null;
			
			/** @brief Contient le format des dates. "M j H:i:s.u" par défaut. */			
			private $dateformat;
			
			/** @brief Contient la liste des listeners, objets qui souhaitent attendrent des événements. */
			private $eventListener=array();
			
			/** @brief Contient le level minimum que l'eventlistener correspondant souhaite recevoir. */
			private $eventLevel=array();
		
			/**
			 * @brief Initialise la variable EventManager::$chrono et configure l'error handler.
			 * @details L'error handler est initialisé pour appeler la méthode EventManager::fire_error_handler()
			 * @param string $defaultdateformat Le format d'affichage des dates. "M j H:i:s.u" par défaut.
			 */
			public function __construct($defaultdateformat="M j H:i:s.u"){
				$this->chrono=microtime(true);
				$this->dateformat=$defaultdateformat;
				set_error_handler(array($this,"fire_error_handler"));
			}
			
			/**
			 * @brief Retourne la date et/ou l'heure actuelle formatée à partir d'un certain format.
			 * @details Supporte les nanosecondes contrairement à date() et à DateTime() mal initialisé.
			 * @see http://php.net/manual/fr/function.date.php pour le formatage de la date.
			 * @param string $format Le format de la date.
			 * @retval string La date formatée avec le format $format.
			 */
			private static function getDateTime($format){
				$t = microtime(true);
				$micro = sprintf("%06d",($t - floor($t)) * 1000000);
				$datetime = new DateTime( date('Y-m-d H:i:s.'.$micro, $t) );
				return $datetime->format($format);
			}
			
			/**
		 	* @brief Retourne le format par défaut des dates générées/utilisées dans l'EventManager
		 	* @retval string Le format des dates 
		 	*/
			public function getDateFormat(){
				return $this->dateformat;
			}
			
			/**
			 * @brief Configure les formats par défaut des dates générées/utilisées dans l'EventManager
			 * @param string $defaultdateformat Le format des dates
			 */
			public function setDateFormat($defaultdateformat){
				$this->dateformat=$defaultdateformat;
			}
			
			/**
			 * @brief Retourne, sous forme de timestamp avec microsecondes, l'heure à laquelle l'objet fut construit.
			 * @retval float Le timestamp avec précision en microsecondes.
			 */
			public function getChronoStart(){
				return $this->chrono;
			}
			
			/**
			 * @brief Retourne, sous forme de timestamp avec microsecondes, l'heure actuelle.
			 * @retval float Le timestamp avec précision en microsecondes.
			 */
			public function getChronoNow(){
				return microtime(true);
			}
			
			/**
			 * @brief Retourne, sous forme de timestamp avec microsecondes, depuis combien de temps l'objet EventManager existe.
			 * @retval float La durée en secondes avec précision en microsecondes.
			*/
			public function getChronoDiff(){
				return $this->getChronoNow() - $this->getChronoStart();
			}
			
			/**
			 * @brief Appelée par l'error_handler afin de transmettre les événements aux listeners.
			 * @details Cette fonction ajoute des informations à $event et envoi cet événénemt aux différents listeners 
			 * d'EventManager::$eventListener qui souhaite reçevoir le type d'erreur référencé par EventManager::$eventLevel.
			 * Les informations ajoutées sont : 
			 * <ul>
			 * <li>datetime   : L'heure à laquelle l'évent à été généré formaté avec EventManager::$dateformat</li>
			 * <li>chronotime : La différence entre la création de l'EventManager et l'heure à laquelle l'event fut généré</li>
			 * </ul>
			 * @param array $event
			 */
			private function fire_error($event){
				$chronotime = sprintf("%.6f",$this->getChronoDiff());
				$time = self::getDateTime($this->dateformat);
				
				$event["datetime"]		= $time;
				$event["chronotime"]	= $chronotime;
				
				foreach($this->eventLevel as $idx => $level){
					if( $event["errno"] <= $level ) call_user_func(array($this->eventListener[$idx],"fireEventReceive"),$event);
				}
			}
			
			/**
			 * @brief Transforme un type d'erreur en errno
			 * @details 
			 * <ul>
			 * 	<li>Une valeur >  0 signifie que $type contient un errno et $type sera donc directement retourné.</li>
			 *  <li>Une valeur <= 0 signifie que $type est utilisé comme un plugin esia précédé de -
			 *  <ul>
			 *  	<li>0 signifie OK et sera replacé par E_USER_NOTICE</li>
			 *  	<li>1 donc -1 signifie WARNING et sera remplacé par E_USER_WARNING</li>
			 *  	<li>2 donc -2 signifie CRITICAL et sera remplacé par E_USER_ERROR</li>
			 *  </ul>
			 * </ul>
			 * @see http://php.net/manual/fr/errorfunc.constants.php pour la liste des constantes errno
			 * @param int $type Un entier >= -2
			 * @return int Un entier > 0 multiple de 2 correspondant à un errno.
			 */
			public static function getErrno($type){
				switch($type){
					case 0  : return E_USER_NOTICE;
					case -1 : return E_USER_WARNING;
					case -2 : return E_USER_ERROR;
					default : return $type;
				}
			}
			
			/**
			 * @brief Méthode de callback que l'EventManager fournit comme error_handler.
			 * @details 
			 * 	Cette méthode est donc appelée à chaque événement généré par php.
			 * 	Transforme ses paramètres en tableau et passe ce tableau comme $event à la methode EventManager::fire_error().
			 * 	Chaque key du tableau correspond à nom du paramètre ($event["errno"] = $errno).
			 * @param int $errno Le niveau d'erreur 
			 * @param string $errstr Le message d'erreur
			 * @param string $errfile Le nom du fichier dans lequel l'erreur a été identifiée. 
			 * @param int $errline Le numéro de la ligne d'ou l'événement est parti.
			 * @param array $errcontext Tableau avec toutes les variables qui existaient lorsque l'erreur a été déclenchée.
			 * @see http://php.net/manual/fr/errorfunc.constants.php pour la liste des constantes errno
			 */
			public function fire_error_handler($errno,$errstr,$errfile,$errline,$errcontext){
				$event = array(
						"errno"			=> $errno,
						"errstr"		=> $errstr,
						"errfile"		=> $errfile,
						"errline"		=> $errline,
						"errcontext"	=> $errcontext,
				);
				
				$this->fire_error($event);
			}
			
			/**
			 * @brief Permet la génération d'un événement à partir d'une Exception. 
			 * @details 
			 * 	Génére un tableau $event sur base de l'exception et passe celui-ci à la méthode EventManager::fire_error().
			 * 	L'event généré contient les valeurs suivantes :
			 *<table>
			 *	<tr><th><b>Nom :</b></th>	<th>Type</th>	<th>Description :</th></tr>
			 *	<tr><td><b>errno</b></td>	<td>int</td>	<td>Le niveau d'erreur</td></tr>
			 *	<tr><td><b>errstr</b></td>	<td>string</td>	<td>Le message d'erreur</td></tr> 
			 *	<tr><td><b>errfile</b></td>	<td>string</td>	<td>Le nom du fichier dans lequel l'erreur a été identifiée.</td></tr> 
			 *	<tr><td><b>errline</b></td>	<td>int</td>	<td>Le numéro de la ligne d'ou l'événement est parti.</td></tr> 
			 *	<tr><td><b>excode</b></td>	<td>int</td>	<td>Le code de l'exception qui a généré l'erreur.</td></tr> 
			 *	<tr><td><b>exsource</b></td><td>string</td>	<td>La source de l'erreur</td></tr> 
			 *</table>
			 * @param Exception $ex L'exception à partir de laquelle on va généré l'event.
			 * @param int $errno Le niveau d'erreur.
			 * @see http://php.net/manual/fr/errorfunc.constants.php pour la liste des constantes errno
			 * @todo Ajouter une description compléte de chacune des clés qu'un événement peut avoir.
			 */
			public function fire_exception(Exception $ex,$errno){
				$event = array(
						"errno"			=> $errno,
						"errstr"		=> $ex->getMessage(),
						"errfile"		=> $ex->getFile(),
						"errline"		=> $ex->getLine(),
						"excode"		=> $ex->getCode(),
				);
				if (is_a($ex,"EsiaException")) $event["exsource"] = $ex->getSource();
			
				$this->fire_error($event);
			}
			
			/**
			 * @brief Génére un événement sur la base d'un string ou d'une exception
			 * @details
			 * <ul>
			 * 	<li>Si $exception est un string, envoi un message à l'error_handler</li>
			 *  <li>Si $exception est une Exception, preveniens directement les listeners via EventManager::fire_exception()</li>
			 * </ul>
			 * Si $type=NULL, $errno vaudra :
			 * <ul>
			 * 	<li>0 (OK) si $exception est un string</li>
			 *  <li>-2 (CRITICAL) si $exception est une Exception</li></li>
			 * </ul>
			 * @param string|Exception $exception Un message sous la forme d'un string ou un Exception
			 * @param int $type Un $type d'erreur.
			 * @see EventManager::getErrno() pour le paramètre $type
			 */
			public function notify($exception,$type=NULL){
				$type_msg=self::getErrno(0);
				$type_ex=self::getErrno(-2);
				
				if($type!=null) $type_msg=$type_ex=self::getErrno($type);
				$message = $exception;
				if(is_string($message)){
					trigger_error($message,$type_msg);
				}
				else if(is_a($message,"Exception")) $this->fire_exception($message,$type_ex);
			}
			
			/**
			 * @brief Ajoute un listener à l'EventManager
			 * @param mixed $listener Un objet possédant une méthode fireEventReceive($event) ou $event est un array.
			 * @param string $level Le level minimum qu'un event doit avoir pour que le listener en soit averti.
			 */
			public function addEventListener($listener,$level=E_USER_NOTICE){
				array_push($this->eventLevel,$level);
				array_push($this->eventListener, $listener);
			}
	}
	
	/**
	 * @todo Documenter
	 * @author Julien George
	 *
	 */
	class stdoutLogger{
		public function __construct(){}
		public function fireEventReceive($event){
			unset($event["errcontext"]);
			
			$errstr = $event["errstr"];
			
			echo "EVENT_RECEIVE : $errstr\n";
		//	print_r($event);
		//	echo "END :\n\n";
		}
	}
	
	/**
	 * @todo Documenter
	 * @author Julien George
	 *
	 */
	class fileLogger{		
		private $path;
		private $append;
		
		public function __construct($file_name,$file_dir="/tmp",$file_append=true){
			$this->path=$file_dir."/".$file_name;
			$this->append=$file_append;
		}
		
		public function fireEventReceive($event){
			$time 	= $event['datetime'];
			$chrono = $event['chronotime'];
			$errno	= $event['errno'];
			$errstr	= $event['errstr'];
			$errline= $event['errline'];
			$errfile= $event['errfile'];
			
			$errno = str_pad($errno, 5);
			
			$code=$source="";
			if( isset($event['excode']))   $code="(".$event["excode"].")";
			if( isset($event['exsource'])) $source="<".$event["exsource"].">";
			
			$exception="$source$code";
			
			$msg ="$time | $chrono | errno : $errno | $errfile:$errline | $exception$errstr\n";
		
			if($this->append) file_put_contents($this->path, $msg,FILE_APPEND);
			else {
				$this->append=true;
				file_put_contents($this->path, $msg);
			}
		}
	}
?>