<?php
/**
        @file
        @author Julien George
        @brief Permet de communiquer via des WebSockets
        @details Les classes utiles de ce fichiers sont WSServerInitiator et WSClientInitiator (Classes filles de WSInitiator) dont le but est de récupérer la classe WSSocket permettant d'envoyer/recevoir des données.
*/

require_once("Exceptions.php");


/**
        @brief Socket connectée de type TCP/SSL.
        @details Cette classe encapsule un Stream de type TCP ou SSL. Par défaut, la socket est blocante.
*/

class SocketStream extends Threaded
{
	/** @brief Utilisé pour définir si la socket doit être en mode blocant ou pas */
	protected $blocking_mode=TRUE;

	/** @brief 2 valeurs possibles : tcp ou ssl */
	protected $context_type;
	/** @brief Le context du stream manipulé dans cette classe */
	protected $context;
	/** @brief Le stream utilisé pour communiquer */
	protected $socket=FALSE;

	/**
		@brief Aide à convertir un binaire en string de 0 et 1.
		@param string $string Le Binaire à convertir
		@return Un string qui est une série de 0 et 1.
	*/
	public static function sprintB($string) {
    		$return = '';
    		for ($i = 0; $i < strlen($string); $i++) $return .= sprintf("%08b", ord($string[$i]));
    		return $return;
  	}

	/**
		@brief Initialise les variables et le context de la socket.
		@details Pour une socket SSL, initialiser la variable $ssl à TRUE et utiliser plusieurs fois la méthode setContextOption($option_name,$value) pour paramètrer le SSL.
		@param boolean $ssl La socketstream sera de type TCP si $ssl vaut FALSE ou de type SSL si $ssl vaut TRUE. 
	*/	
	public function __construct($ssl=false)
	{
			if( $ssl ) 
			{
				$context_type="ssl";
			}
			else 
			{
				$this->url = $context_type="tcp";
			}
			
			$this->context_type=$context_type;
			$this->context=stream_context_create();
	}

	/**
		@brief Configure les paramètres pour la socketstream TCP/SSL.
		@details A utiliser avant un connect ou un bind. L'utilisation de cette méthode est surtout utile, obligatoire même quand on souhaite faire du SSL.
		@param string $option_name Le nom de l'option à configurer.
		@param string $value La valeur de l'option à configurer.
		@see Création de sockets SSL en PHP : http://www.devdungeon.com/content/how-use-ssl-sockets-php
		@see Liste des options SSL : http://php.net/manual/fr/context.ssl.php
		@see Liste des options TCP : http://php.net/manual/fr/context.ftp.php
	*/
	public function setContextOption($option_name,$value){
		return stream_context_set_option($this->context,$this->context_type,$option_name,$value);
	}

	/**
		@brief Initialisation du mode blocant ou pas. 
		@details Utilisé afin de configurer le mode blocant ou pas du stream en fonction du contenu de la propriété SocketStream::$blocking_mode
	*/
	private function initBlockingMode(){
		if(! stream_set_blocking ( $this->socket, $this->blocking_mode )) throw new Exception("Unable to configure socket on blocking mode => ".$this->blocking_mode ? "TRUE" : "FALSE"  );
	}

	/**
		@brief Configure si la stream doit être blocante. 
		@details Set la propriété SocketStream::$blocking_mode. Si le stream SocketStream::$socket existe, il est setté, sinon, il sera setté à sa création.
		@param boolean $mode TRUE ou FALSE si la socket doit être bloquante ou pas
		@see Ce mode bloquant ou pas modifie le comportment de SocketStream::receive($length,$force=false)
	*/
	public function setBlockingMode($mode){
		if( $this->blocking_mode != $mode ){
			$this->blocking_mode = $mode;
			if( $this->isConnected() ) $this->initBlockingMode();
		}
	}

	/**
		@brief Retourne l'état blocant ou pas de SocketStream
		@retval boolean TRUE ou FALSE si la socket est bloquante ou pas
	*/
	public function isBlockingMode(){
		return $this->blocking_mode;
	}

	/**
		@brief Retourne l'état encrypted de la socket
		@retval boolean TRUE ou FALSE si la socket est SSL ou TCP
	*/
	public function isEncrypted(){
		return ($this->context_type == "ssl");
	}

	/**
		@brief Indique si la StreamSocket est connectée
		@retval boolean TRUE ou FALSE si la socket est connectée ou pas
	*/
	public function isConnected(){
		return ( $this->socket !== FALSE && is_resource($this->socket) && !feof($this->socket) ); 
	}

	/**
		@brief Bind le stream sur le port
		@details Crée la socket, la bind sur le port et effectue le listen
		@param string $addr L'adresse IP sur laquelle on se bind.
		@param int $port Le numéro de port
		@exception SocketConnectedException Si la SocketStream est déjà connectée.
		@exception SocketBindException Si le BIND ou le listen du stream à échoué.
	*/
	public function bind($addr,$port){
		if( $this->isConnected() ) throw new SocketConnectedException("The socket is already connected");
		
		
		$url=$this->context_type."://$addr:$port";
		$this->socket = stream_socket_server($url,$errno,$errstr,STREAM_SERVER_BIND|STREAM_SERVER_LISTEN,$this->context);
		if( !$this->isConnected() ) throw new SocketBindException("$errno : $errstr");
		$this->initBlockingMode();
	}

	/**
		@brief Accepte la connection d'un client
		@details Si $timeout n'est pas setté ou vaut null, sa valeur sera identique à celle du "default_socket_timeout" du fichier ini.
		@param int $timeout Le nombre de seconde à attendre pour la connection d'un client.
		@exception SocketClosedException Si la SocketStream est fermée.
		@exception SocketAcceptException Si une erreur s'est produite.
	*/
	public function accept($timeout=null,$notThreaded=false){
		if( !$this->isConnected() ) throw new SocketClosedException("Unable accept connections on closed socket");
		if( $timeout == null ) $timeout=ini_get("default_socket_timeout");

		set_error_handler(function($errno, $errstr, $errfile, $errline, array $errcontext) {
			// error was suppressed with the @-operator
			$ignore=array(0,E_USER_DEPRECATED,E_USER_ERROR,E_USER_NOTICE,E_USER_WARNING);
			if(in_array($errno,$ignore))  { return false; }
			throw new SocketConnectException($errstr);
		});
		$socket_client = stream_socket_accept($this->socket,$timeout);
		restore_error_handler();
		if($notThreaded===false)
		{
			$ret=new SocketStream($this->isEncrypted());
		}
		else
		{
			$ret=new SocketStreamNotThreaded($this->isEncrypted());
		}



		$ret->socket=$socket_client;
		return $ret;
	}

	/**
		@brief Connecte la StreamSocket à un serveur. 
		@details Si $timeout n'est pas setté ou vaut null, sa valeur sera identique à celle du "default_socket_timeout" du fichier ini.
		@param string $addr L'adresse IP ou le hostname du serveur
		@param int $port Le numéro de port du serveur
		@param int $timeout Le nombre de seconde à attendre avant que la connection échoue.
		@exception SocketClosedException Si la SocketStream est fermée.
		@exception SocketConnectException Si une erreur s'est produite lors de la connection du client.
	*/
	public function connect($addr,$port,$timeout=null){
		if( $this->isConnected() ) throw new SocketConnectedException("The socket is already connected");
		if( $timeout == null ) $timeout=ini_get("default_socket_timeout");

		$url=$this->context_type."://$addr:$port";

		set_error_handler(function($errno, $errstr, $errfile, $errline, array $errcontext) {
    		// error was suppressed with the @-operator
    		if (0 === error_reporting()) { return false; }
			throw new SocketConnectException($errstr);
		});

		$this->socket = stream_socket_client($url,$errno,$errstr,$timeout,STREAM_CLIENT_CONNECT,$this->context);
		restore_error_handler();
	} 

	/**
		@brief Retourne l'adresse IP et le port de l'hôte distant connecté.
		@retval string L'adresse de l'hôte distant sous forme \<string\>:\<port\>
		@exception SocketClosedException Si la SocketStream est fermée.
	*/
	public function getPeerAddr(){
		if( !$this->isConnected() ) throw new SocketClosedException("Unable to get peer addr of closed socket");
		return stream_socket_get_name($this->socket,true);
	}

	/**
		@brief Retourne l'adresse IP et le port local.
		@retval string L'adresse locale sous forme \<string\>:\<port\>
		@exception SocketClosedException Si la SocketStream est fermée.
	*/
	public function getLocalAddr(){
		if( !$this->isConnected() ) throw new SocketClosedException("Unable to get local addr of closed socket");
		return stream_socket_get_name($this->socket,false);
	}

	/**
		@brief Envoi des données vers l'hôte distant.
		@param raw $data les données à envoyer.
		@exception SocketClosedException Si la SocketStream est fermée.
		@exception SocketWriteException Si l'écriture des données à échouer.
	*/
	public function send($data){
		if( !$this->isConnected() ) throw new SocketClosedException("Unable write on closed socket");
		
		$written = fwrite($this->socket, $data, strlen($data));
		if ($written === FALSE)
                        throw new SocketWriteException("Could only write $written out of " . strlen($data) . " bytes".socket_strerror(socket_last_error()));
		$data=null;
	}

	/**
		@brief Se met en attente d'un changement sur le stream.
		@details Se bloque jusqu'à l'arrivée du $timeout, d'une exception ou de données sur le stream.
		@param int $timeout le nombre de secondes à attendre ou 0 pour infini.
		@exception SocketClosedException Si la SocketStream est fermée.
	*/
	public function waitData($timeout=0){
		if( !$this->isConnected() ) throw new SocketClosedException("Unable too wait on closed socket");
		$sockets=array($this->socket);
		$nullempty=array();
		return stream_select($sockets,$nullempty,$sockets,$timeout);
	}

	/**
		@brief Reçoit des données vers l'hôte distant.
		@details Le comportement dépends du mode bloquant et du param $force
			<ul>
				<li>Mode BLOQUANT = TRUE  :
					<ul>
						<li>Param $force = FALSE : Bloque jusqu'à recevoir un octet minimum et retourne entre 1 et $length octet.</li>
						<li>Param $force = TRUE : Bloque jusqu'à recevoir $lenght octets.</li>
					</ul>
				</li>
				<li>Mode BLOQUANT = FALSE  :
					<ul>
						<li>Param $force = FALSE : Retourne  0 octets si le buffer est vide</li>
						<li>Param $force = TRUE : Bloque jusqu'à recevoir $lenght octets. Pour des raison de performances il est conseillé d'utiliser en mode BLOQUANT</li>
					</ul>
	
				</li>
			</ul>
		@param int $length La quantité maximum de données à lire (en octets).
		@param boolean $force Si TRUE, boucle sur la stream jusqu'à avoir reçut $lenght octets. Déconseillé en Mode non BLOQUANT
		@return Les données reçues.
		@exception SocketClosedException Si la SocketStream est fermée.
		@exception SocketReadException Si l'écriture des données à échouer.
		@see SocketStream::setBlockingMode($mode)
	*/
	public function receive($length,$force=false){
		if( !$this->isConnected() ) throw new SocketClosedException("Unable too read on closed socket");
		if($force){
			$data="";
			$actual_read=0;

			$data=$data.$this->receive($length,false);
			$actual_read=strlen($data);

			while($actual_read<$length){
				$this->waitData();
				$data=$data.$this->receive($length-$actual_read,false);
				$actual_read=strlen($data);
			}
		}
		else{
			$data=@fread($this->socket,$length);
			if( $data === FALSE ) 
				throw new SocketReadException("Unable to read on socket");
		}
		return $data;
	}
	
	/**
	 * @brief Permet d'obtenir la resource utilisée par la SocketStream afin de communiquer via TCP.
	 * @retval boolean|resource Retourne la resource ou false si il n'y en à pas
	 */
	public function getResource(){
		return $this->socket;
	}
	
	/**
	 * @brief Permet d'initialiser la resource utilisée par la SocketStream afin de communiquer via TCP.
	 * @param resource $resource La resource permettant la communication réseau.
	 */
	public function setResource($resource){
		$this->socket=$resource;
	}

	/**
		@brief Ferme le stream si celui-ci est connecté.
	*/
        public function close(){
                if( $this->socket != null && is_resource($this->socket) ) {
                        fflush( $this->socket );
                        stream_socket_shutdown ( $this->socket, STREAM_SHUT_RDWR );
                        fclose ( $this->socket );
                }
                unset($this->context);
                unset($this->socket);
                $this->socket=null;
        }

}

/**
		@class WSSocket
        @brief Socket de communication en WebSocket.
        @details Permet d'envoyer/recevoir des données à travers une WebSocket connectée (TCP ou SSL). Manipule un objet SocketStream.
		@example ws_callback_receive.php Fonction de callback acceptée par la classe WSSocket.
*/
class WSSocket extends Threaded{
	/** 
		@brief Enumére les différents opcodes possibles pour un fragment WebSocket
		@details
		<table>
			<tr>
				<th><b>Nom :</b></th>
				<th>Description :</th>
			</tr>
			<tr>
				<td><b>continuation</b></td>	<td>Indique que c'est la suite d'une donnée fragmentée.</td></tr> 
			<tr>
				<td><b>text</b></td>			<td>Indique que la donnée est de type text.</td>
			</tr> 
			<tr>
				<td><b>binary</b></td>			<td>Indique que la donnée est de type binary.</td>
			</tr> 
			<tr>
				<td><b>close</b></td>			<td>Annonçe une demande de fermeture de connection ou la réponse à celle-ci.</td>
			</tr> 
			<tr>
				<td><b>ping</b></td>			<td>Permet de pinguer l'hôte distant.</td>
			</tr> 
			<tr>
				<td><b>pong</b></td>			<td>Réponse au ping.</td>
			</tr> 
		</table>
		@remark
		<ul>
			<li>L'opcode "continuation" n'est utilisé que si les données sont fragmentées.</li>
			<li>Lors d'une fragmentation, le premier fragment contient l'opcode "text|binary", les autres contiendront "continuation".</li>
			<li>Les opcodes "close|ping|pong" ne devrait jamais être fragmentés car la taille de leurs données ne doit pas excédé 125 octets.</li>
		</ul>
	*/
    const OPCODES = array(
                    'continuation' => 0,
                    'text'         => 1,
                    'binary'       => 2,
                    'close'        => 8,
                    'ping'         => 9,
                    'pong'         => 10,
    );

	/** @brief Nom du protocol de communication WebSocket choisi lors du handshake WebSocket */
	protected $protocol_receiver;
	
	/** @brief Un Objet de type SocketStream utilisé pour envoyé/recevoir des données bruts */
	protected $socket;

	/** @brief Taille maximum qu'un fragment websocket peut avoir. Si la donnée dépasse cette taille, elle sera fragmentée. */
	protected $fragmentSize;	

	/** @brief Mode de blocage par défaut de la socket. */
	protected $default_blocking_mode;

	/** @brief Mode de blocage actual de la socket. */
	protected $actual_blocking_mode;
	
	/** @brief Permet à la méthode WSSocket::receive_fragment() de garder en mémoire les données déjà reçue dans le cadre d'une fragmentation de données */
	protected $huge_payload=null;
	
	/** @brief Permet à la méthode WSSocket::receive_fragment() de se rappeler si il est en cours de fermeture ou pas */
	protected $is_closing=false;
	
	/** @brief Permet à la méthode WSSocket::receive_fragment() de connaître le premier opcode reçut dans une fragmentation de données */
	protected $last_opcode=null;
	/**
		@brief Initialise la WebSocket avec son WSSocket::$protocol_receiver, sa SocketStream et sont WSSocket::$fragmentSize.
		@param string $protocol Le protocol utilisé au dessus de la WebSocket.
		@param SocketStream $socket Stream TCP/SSL qui à déjà réalisé un Handshake WebSocket.
		@param int $fragment_size (Default = 1300) La taille maximum que pourra faire un fragment WebSocket. 
	*/
	public function __construct($protocol,SocketStream $socket,$fragment_size=2300){
		$this->protocol_receiver=$protocol;	
		$this->socket=$socket;
		$this->default_blocking_mode=$this->actual_blocking_mode=$socket->isBlockingMode();
		$this->fragmentSize=$fragment_size;
	}

	/**
		@brief Configure le WSSocket::$fragmentSize.
		@details Si la taille d'une donnée envoyée est supérieure à $size, alors cette donnée sera fragmentée en fragments de 1300 octets (excepté le dernier). 
		@param int $size Taille maximum que peut avoir un fragment WebSocket.
	*/
	public function setFragmentSize($size){
		$this->fragmentSize = $size;
	}

	/**
		@brief Retourne le WSSocket::$fragmentSize.
		@details Si la taille d'une donnée envoyée est supérieure à $size, alors cette donnée sera fragmentée en fragments de 1300 octets (excepté le dernier). 
		@retval int $size Taille maximum que peut avoir un fragment WebSocket.
	*/
	public function getFragmentSize(){
		return $this->fragmentSize;
	}

	/**
		@brief Retourne TRUE si la StreamSocket est actuellement en mode blocant.
		@retval boolean TRUE ou FALSE si la StreamSocket est en mode blocant ou pas.
		@see SocketStream::receive() pour plus d'informations.
	*/
	private function isBlockingMode(){
		return $this->actual_blocking_mode;
	}

	/**
		@brief Configure le mode blocant de la StreamSocket.
		@param boolean $mode TRUE ou FALSE si on souhaite avoir la StreamSocket en mode blocant ou pas.
		@see SocketStream::receive() pour plus d'informations.
	*/
	private function setBlockingMode($mode){
		if( $this->actual_blocking_mode != $mode ){
			$this->actual_blocking_mode = $mode;
			$this->socket->setBlockingMode($mode);
		} 
	}

	/**
		@brief Retourne TRUE si le mode par défaut de la StreamSocket est actuellement en mode blocant.
		@retval boolean TRUE ou FALSE si la StreamSocket est par défaut en mode blocant ou pas.
		@see SocketStream::receive() pour plus d'informations.
	*/
	public function getDefaultBlockingMode(){
		return $this->default_blocking_mode;
	}

	/**
		@brief Configure le mode blocant par défaut de la StreamSocket.
		@param boolean $mode TRUE ou FALSE si on souhaite configurer la valeur par défaut de la StreamSocket en mode blocant ou pas.
						TRUE =  mode bloquant
		@see SocketStream::receive() pour plus d'informations.
	*/
	public function setDefaultBlockingMode($mode){
		if( $this->default_blocking_mode != $mode ){
			$this->default_blocking_mode = $mode;
			$this->setBlockingMode($mode);
		}
	}

	/**
		@brief Retourne le protocole applicatif utilisé au dessus de la WebSocket.
		@retval string Le nom du protocole applicatif.
	*/
	public function getProtocol(){
		return $this->protocol_receiver;
	}

	/**
		@brief Retourne l'adresse IP et le port de l'hôte distant connecté.
		@retval string L'adresse de l'hôte distant sous forme \<string\>:\<port\>
		@exception SocketClosedException Si la SocketStream est fermée.
	*/
	public function getPeerAddr(){
		return $this->socket->getPeerAddr();
	}

	/**
		@brief Retourne l'adresse IP et le port local.
		@retval string L'adresse locale sous forme \<string\>:\<port\>
		@exception SocketClosedException Si la SocketStream est fermée.
	*/
	public function getLocalAddr(){
		return $this->socket->getLocalAddr();
	}

	/**
		@brief Retourne l'état encrypted de la socket
		@retval boolean TRUE ou FALSE si la socket est SSL ou TCP
	*/
	public function isEncrypted(){
		return $this->socket->isEncrypted();
	}

	/**
		@brief Indique si la StreamSocket est connectée
		@retval boolean TRUE ou FALSE si la socket est connectée ou pas
	*/
	public function isConnected(){
		return $this->socket->isConnected();
	}	

	/**
		@brief Envoit un fragment WebSocket.
		@details 
		@param boolean $final Indique si c'est le dernier fragment. 
		@param raw $payload	La donnée à envoyer.
		@param string $opcode Un WSSocket::OPCODES.
		@param boolean $masked Indique si il faut masquer ou pas les données. 
		@warning $masked=true Ne veut pas dire sécurisé. Il est conseillée de l'avoir à false (pour les débuggages) et utiliser SSL pour la sécurité.
	*/
	private function send_fragment($final, $payload, $opcode,$masked=true){
		$frame_head_binstr = '';
		$frame_head_binstr .= (bool) $final ? '1' : '0';		//Byte FIN
		$frame_head_binstr .= '000';					//Bytes RSV 1-3
        $frame_head_binstr .= sprintf('%04b', WSSocket::OPCODES[$opcode]);        //Opcode 
		$frame_head_binstr .= $masked ? '1' : '0';			//Masked ?
	
		//Taile du payload	
		$payload_length = strlen($payload);
    		if ($payload_length > 65535) {
      			$frame_head_binstr .= decbin(127);
      			$frame_head_binstr .= sprintf('%064b', $payload_length);
    		}
    		elseif ($payload_length > 125) {
      			$frame_head_binstr .= decbin(126);
      			$frame_head_binstr .= sprintf('%016b', $payload_length);
    		}
    		else {
      			$frame_head_binstr .= sprintf('%07b', $payload_length);
    		}
		
		$frame = '';
		// Write frame head to frame.
		foreach (str_split($frame_head_binstr, 8) as $binstr) $frame .= chr(bindec($binstr));
	
		// Handle masking
    		if ($masked) {
      			// generate a random mask:
      			$mask = '';
      			for ($i = 0; $i < 4; $i++) $mask .= chr(rand(0, 255));
      			$frame .= $mask;
    		}		
		
		// Append payload to frame:
    		for ($i = 0; $i < $payload_length; $i++) {
      			$frame .= ($masked === true) ? $payload[$i] ^ $mask[$i % 4] : $payload[$i];
    		}

		$this->socket->send($frame);
	}

	/**
		@brief Appelé par la fonction WSSocket::receive_fragment() afin d'avertir qu'un fragment est arrivé.
		@param null|string|array $callback_function
			<p>La ou les fonctions de callback (Voir l'example ws_callback_receive.php). Ses functions peuvent être passées sous les formes suivantes :</p>
			<ul>
				<li>null : Aucune function de callback donc la function ne fera rien du tout.</li>
				<li>string : Le string correspondant à une fonction de callback qui sera appelée quelque soit l'opcode de la donnée reçue.</li>
				<li>array\<int,string\> : Cette simple liste contient des méthodes qui seront toutes appelées quelque soit le opcode reçu.</li>
				<li>array(object,string)></li> : Liste de type (Object => methode), La méthode doit appartenir à l'Object. Toutes ces méthodes seront appelées quelque soit l'opcode reçu.
				<li>array\<string,string\> : Liste (opcode => function) contenant la fonction à appeler en fonction de l'opcode reçu</li>
			</ul>
		@param string $opcode L'opcode de la donnée, de type WSSocket::OPCODES.
		@param raw $data La donnée reçue.
		@param null|int $close_status Le status du close. Toujours accompagné de l'opcode "close".
	*/
  	private function fireReceive($callback_function,$opcode,$data,$close_status=null){
                if(is_string($callback_function)) {
                        call_user_func($callback_function,$this,$opcode,$data,$close_status);
                        return;
                }
                if( is_array($callback_function) )
                {
                        if(is_object($callback_function[0]))
                        {
                                call_user_func(array($callback_function[0], $callback_function[1]),$this,$opcode,$data,$close_status);
                        }
                        else
                        {
                                foreach( $callback_function as $key => $value)
                                {
                                        if( is_int($key) ) call_user_func($callback_function,$this,$opcode,$data,$close_status);
                                        else
                                        {
                                                if( $key == $opcode ) {
                                                        call_user_func($callback_function,$this,$opcode,$data,$close_status);
                                                        return;
                                                }
                                        }
                                }
                        }
                }
                return;
     }

	/** 
		@brief Permet de recevoir un fragment WebSocket
		@details
			<ul>
			<li>Lit une donnée dans le buffer et appelle ensuite la méthode WSSocket::fireReceive() afin de prévenir qu'une donnée à été lue.</li>
			<li>Si le buffer est vide et que la $wait=false alors aucune méthode de callback ne sera appelée.</li>
			<li>Lorsque la donnée est incompléte (Flag Final à false), la méthode WSSocket::fireReceive() est appelée uniquement si $all_fragments=true. Cette méthode sera alors appelée avec $data=null.</li>
			<li>Cette méthode réponds automatiquement à un close (si c'est le peer qui à envoyé le premier close) par un close et à un ping par un pong.</li>
			</ul>
		@param null|string|array $callback_function
			<p>La ou les fonctions de callback (Voir l'example ws_callback_receive.php). Ses functions peuvent être passées sous les formes suivantes :</p>
			<ul>
				<li>null : Aucune function de callback, la methode lira un fragment entier et l'ignorera</li>
				<li>string : Le string correspondant à une fonction de callback qui sera appelée quelque soit l'opcode de la donnée reçue.</li>
				<li>array\<int,string\> : Cette simple liste contient des méthodes qui seront toutes appelées quelque soit le opcode reçu.</li>
				<li>array\<object,string\></li> : Liste de type (Object => methode), La méthode doit appartenir à l'Object. Toutes ces méthodes seront appelées quelque soit l'opcode reçu.
				<li>array\<string,string\> : Liste (opcode => function) contenant la fonction à appeler en fonction de l'opcode reçu</li>
			</ul>
		@param boolean $wait Indique si il faut attendre ou pas qu'un fragment arrive dans le Buffer.
		@param boolean $all_fragments Indique si il faut notifier avec la méthode WSSocket::fireReceive() à chaque fragment ou uniquement à l'arrivée du dernier fragment.
		@retval string
			<ul>
				<li>NODATA : La méthode a été appelée avec $wait=false et aucun fragment n'était présent dans le buffer à ce moment la</li>
				<li>FINAL  : Le fragment lut avait le flag final et donc aucun fragment ne va suivre. La donnée est compléte</li>
				<li>FRAGMENT : D'autres fragments vont suivres. La donnée n'est pa sencore compléte.</li>
			</ul>
	*/
	private function receive_fragment($callback_function,$wait=false,$all_fragments=false){
		//Lecture non blocante des 2 premiers octets
		$data=$this->socket->receive(1,$wait);
		if( strlen($data) == 0 ) return "NODATA";
		$data=$data.$this->socket->receive(1,false);

		//Parsing des 2 premiers octets
		$final = (boolean) (ord($data[0]) & 1 << 7);
		$rsv1  = (boolean) (ord($data[0]) & 1 << 6);
    	$rsv2  = (boolean) (ord($data[0]) & 1 << 5);
    	$rsv3  = (boolean) (ord($data[0]) & 1 << 4);	
		$opcode_int = ord($data[0]) & 31; // Bits 4-7
		$opcode_ints = array_flip(WSSocket::OPCODES);
    	if (!array_key_exists($opcode_int, $opcode_ints)) {
      		throw new Exception("Bad opcode in websocket frame: $opcode_int");
    	}

		$opcode = $opcode_ints[$opcode_int];
		
		$mask = (boolean) (ord($data[1]) >> 7);  // Bit 0 in byte 1

		//Calcul de la taille
		$payload_length = (integer) ord($data[1]) & 127; // Bits 1-7 in byte 1
    		if ($payload_length > 125) {
      			if ($payload_length === 126) $data = $this->socket->receive(2,true); // 126: Payload is a 16-bit unsigned int
      			else                         $data = $this->socket->receive(8,true); // 127: Payload is a 64-bit unsigned int
      			$payload_length = bindec(SocketStream::sprintB($data));
    		}
		
		if ($mask) $masking_key = $this->socket->receive(4,true);
		
		
    		// Get the actual payload, if any (might not be for e.g. close frames.
    	if ($payload_length > 0) {
			$data=$this->socket->receive($payload_length,true);
      		if ($mask) {
        	// Unmask payload.
        		for ($i = 0; $i < $payload_length; $i++) $payload .= ($data[$i] ^ $masking_key[$i % 4]);
      		}
      		else $payload = $data;
    	}
    	
		// if this is not the last fragment, then we need to save the paylo
    	if (!$final) {
      			$this->huge_payload .= $payload;
    	}
		else if ($this->huge_payload) {
			  // sp we need to retreive the whole payload
      			$payload = $this->huge_payload .= $payload;
      			$this->huge_payload = null;
		}
		
		switch($opcode){
			case "text" :
			case "binary" :
				$this->last_opcode=$opcode;
			case "continuation" :
				if($final)				$this->fireReceive($callback_function,$this->last_opcode,$payload,null);
				else if($all_fragments)	$this->fireReceive($callback_function,$this->last_opcode,null,null);
				break;

			case "close" :
				if ($payload_length >= 2) {
					$status_bin = $payload[0] . $payload[1];
        			$status = bindec(sprintf("%08b%08b", ord($payload[0]), ord($payload[1])));
        			$payload = substr($payload, 2);
				
					if (!$this->is_closing) {
						$this->send_fragment(true,$status_bin . 'Close acknowledged: ' . $status, 'close');
						$this->fireReceive($callback_function,$opcode,$payload,$status);
					}
					else {
						$this->fireReceive($callback_function,$opcode,$payload,$status);
					}
					$this->socket->close();
				}
				break;
			case "ping" :
				$this->fireReceive($callback_function,$opcode,$data,null);
				$this->send_fragment(true,$payload,"pong");
				break;
			case "pong" :
				$this->fireReceive($callback_function,$opcode,$data,null);
				break;
		}
		if($final) $this->last_opcode=null;
		
		if($final) return"FINAL";
		else return "FRAGMENT";
	}

	/** 
		@brief Permet de recevoir une donnée via une WebSocket
		@details Boucle sur la méthode WSSocket::receive_fragment() jusqu à avoir lut toute une donnée. WSSocket::fireReceive() sera appelé au dernier fragment sauf si $all_fragments=true.
		@param null|string|array $callback_function
			<p>La ou les fonctions de callback (Voir l'example ws_callback_receive.php). Ses functions peuvent être passées sous les formes suivantes :</p>
			<ul>
				<li>null : Aucune function de callback, la methode lira un fragment entier et l'ignorera</li>
				<li>string : Le string correspondant à une fonction de callback qui sera appelée quelque soit l'opcode de la donnée reçue.</li>
				<li>array\<int,string\> : Cette simple liste contient des méthodes qui seront toutes appelées quelque soit le opcode reçu.</li>
				<li>array\<object,string\></li> : Liste de type (Object => methode), La méthode doit appartenir à l'Object. Toutes ces méthodes seront appelées quelque soit l'opcode reçu.
				<li>array\<string,string\> : Liste (opcode => function) contenant la fonction à appeler en fonction de l'opcode reçu</li>
			</ul>
		@param null|boolean $wait Indique si on souhaite le mode blocant ou pas. null signifie que se sera la méthode WSSocket::getDefaultBlockingMode() qui en décidera.
		@param boolean $all_fragments Indique si il faut notifier avec la méthode WSSocket::fireReceive() à chaque fragment ou uniquement à l'arrivée du dernier fragment.
		@retval boolean TRUE si une donnée à été lue en entiere et false dans le cas contraire.
	*/
	public function receive($callback_function=null,$wait=null,$all_fragments=false){
		//Configuration du mode blockant si nécessaire
		if($wait == null ) $mode=$this->getDefaultBlockingMode();
		else {
			$mode = $wait;
			$this->setBlockingMode($mode);
		}

		//Réception des données
		do{
			$final=$this->receive_fragment($callback_function,$mode,$all_fragments);				
		}while($final=="FRAGMENT");

		//Si mode blocant modifié, remise en configuration par défaut
		if($wait != null ) $this->setBlockingMode($this->getDefaultBlockingMode());
		
		//Valeur de retour
		if( $final == "NODATA" ) return false;
		return true;
	}
	/**
		@brief Envoi une donnée à travers la WebSocket.
		@details Fragmente la donnée si celle-ci est plus grande que WebSocket::fragmentSize
		@param raw $text Les données à envoyer.
		@param string $opcode L'opcode du fragment, doit $etre de type WSSocket::OPCODES. L'opcode "continue" n'est pas accepté car la fragmentation est géré dans la WSSocket
		@exception WebSocketIncorrectOpcodeException Si l'opcode passé est incorrect.
		@exception WebSocketSizeException Si l'opcode close, ping ou pong est envoyé avec une donnée > 125 octets. text et binary n'ayant pas de limite.
	*/
	public function send($text,$opcode="text"){
		$payload_length=strlen($text);

		if( !in_array($opcode,WSSocket::OPCODES) )
			throw new WebSocketIncorrectOpcodeException("The opcode $opcode doesn't exist");
		if( $opcode == "continue" )
			throw new WebSocketIncorrectOpcodeException("You can't use the continue opcode");
		if( $payload_length >= 125 && ( $opcode != "text" && $opcode != "binary" ) )
			throw new WebSocketSizeException("The $opcode opcodes type must have data size <= 125. This payload_length == $payload_length");
	
		//La boucle suivante fragment et envoi le texte 
		$fragment_cursor = 0;
		while( $payload_length > $fragment_cursor ){
			$sub_payload = substr($text, $fragment_cursor, $this->fragmentSize);
			$fragment_cursor += $this->fragmentSize;
			$final = $payload_length <= $fragment_cursor;

			$this->send_fragment($final, $sub_payload, $opcode);
			$opcode = 'continuation';
		}	
	}

	/**
		@brief Envoi un ping et attends une réponse pour la transmettre en callback.
		@details Cette réponse pourrait bien être un pong.
		@param null|string|array $callback_function
			<p>La ou les fonctions de callback (Voir l'example ws_callback_receive.php). Ses functions peuvent être passées sous les formes suivantes :</p>
			<ul>
				<li>null : Aucune function de callback, la methode lira un fragment entier et l'ignorera</li>
				<li>string : Le string correspondant à une fonction de callback qui sera appelée quelque soit l'opcode de la donnée reçue.</li>
				<li>array\<int,string\> : Cette simple liste contient des méthodes qui seront toutes appelées quelque soit le opcode reçu.</li>
				<li>array\<object,string\></li> : Liste de type (Object => methode), La méthode doit appartenir à l'Object. Toutes ces méthodes seront appelées quelque soit l'opcode reçu.
				<li>array\<string,string\> : Liste (opcode => function) contenant la fonction à appeler en fonction de l'opcode reçu</li>
			</ul>
		@param $payload La donnée à envoyer dans le ping. Elle sera retransmise dans le pong. 
	*/
	public function ping($callback_function,$payload){
		$this->send($payload,"ping");
		$this->receive($callback_function);
	}

	/**
		@brief Envoi un close et attends une confirmation puis ferme la SocketStream.
		@details Envoi à la function de callback le status de la réponse close avec les données.
		@param null|string|array $callback_function
			<p>La ou les fonctions de callback (Voir l'example ws_callback_receive.php). Ses functions peuvent être passées sous les formes suivantes :</p>
			<ul>
				<li>null : Aucune function de callback, la methode lira un fragment entier et l'ignorera</li>
				<li>string : Le string correspondant à une fonction de callback qui sera appelée quelque soit l'opcode de la donnée reçue.</li>
				<li>array\<int,string\> : Cette simple liste contient des méthodes qui seront toutes appelées quelque soit le opcode reçu.</li>
				<li>array\<object,string\></li> : Liste de type (Object => methode), La méthode doit appartenir à l'Object. Toutes ces méthodes seront appelées quelque soit l'opcode reçu.
				<li>array\<string,string\> : Liste (opcode => function) contenant la fonction à appeler en fonction de l'opcode reçu</li>
			</ul>
		@param $status Le status du close. 
		@param $message La donnée à envoyer dans le close. Elle sera retransmise dans la confirmation du close. 
	*/
	public function close($callback_function,$status=1000,$message="close"){
		$status_binstr = sprintf('%016b', $status);
		$status_str = '';
		foreach (str_split($status_binstr, 8) as $binstr) $status_str .= chr(bindec($binstr));
		$this->is_closing = true;
		$this->send($status_str . $message, 'close');
		$this->receive($callback_function);
		$this->socket->close();
		$this->socket->isGarbage();
	}
	/**
	 * @brief détruit la socket
	 */
	public function destroy()
	{
		$this->socket->close();
	}
	
	/**
	 * @brief Retourne une objet WSRessourceNoThreaded.
	 * @details Permet d'obtenir un objet stockable dans un autre objet Threaded (sans activer l'"immutabilité" et reconstruire un WSSocket para après sans perdre la connection WS. 
	 * @retval WSResourceNoThreaded Les données nécessaire pour la reconstruction de la WSSocket.
	 */
	public function getNoThreaded(){
		return new WSResourceNoThreaded($this->socket->getResource(),$this->protocol_receiver,$this->fragmentSize,$this->default_blocking_mode);
	}	
	
	/**
	 * @brief Retourne un WSSocket à partir d'un objet WSResourceNoThreaded
	 * @param WSResourceNoThreaded $noThreaded Un objet no Threaded donc sanss problème d'immutabilité dans un autre objet Threaded.
	 * @retval WSSocket Un objet permettant de lire ou d'écrire sur la WebSocket.
	 */
	public static function getFromNoThreaded(WSResourceNoThreaded $noThreaded){
		$socketStream = new SocketStream();
		$socketStream->setResource($noThreaded->getResource());
		
		$wssocket = new WSSocket($noThreaded->getProtocolReceiver(),$socketStream,$noThreaded->getFragmentSize());
		$wssocket->setBlockingMode($noThreaded->getBlockingMode());
		return $wssocket;
	}
}

/**
 * @author Julien George
 * @brief No threaded ressource et paramètres pour une WSSocket
 * @details Permet de stocker la resource et les paramètres nécessaires à la création d'une WSSocket. Sauf que cet objet peut être stocké dans un Objet Treaded sans causé l'immutabilité. 
 *
 */
class  WSResourceNoThreaded{
	/** @brief Le stream utilisé pour communiquer */
	protected $resource;
	
	/** @brief Nom du protocol de communication WebSocket choisi lors du handshake WebSocket */
	protected $protocol_receiver;
	
	/** @brief Taille maximum qu'un fragment websocket peut avoir. Si la donnée dépasse cette taille, elle sera fragmentée. */
	protected $fragment_size;
	
	/** @brief Mode de blocage par défaut de la socket. */
	protected $default_blocking_mode;
	
	/**
	 * Le constructeur contentant la resource et les paramètres à stocker.
	 * @param resource $resource
	 * @param string $protocol
	 * @param int $fragment
	 * @param boolean $blocking_mode
	 */
	public  function __construct($resource,$protocol,$fragment,$blocking_mode){
		$this->resource=$resource;
		$this->protocol_receiver=$protocol;
		$this->fragment_size=$fragment;
		$this->default_blocking_mode=$blocking_mode;
	}
	
	/**
	 * @brief Retourne la resource permettant la communication WebSocket
	 * @retval resource La resource
	 */
	public function getResource(){
		return $this->resource;
	}
	
	/**
	 * @brief Retourne le nom du protocol
	 * @retval string le protocol
	 */
	public function getProtocolReceiver(){
		return $this->protocol_receiver;
	}
	
	/**
	 * @brief Retourne la taille d'un fragment WSSocket
	 * @retval int La taille du fragment
	 */
	public function getFragmentSize(){
		return $this->fragment_size;
	}
	
	/**
	 * @brief Retourne si la socket est blocante ou pas
	 * @retval boolean mode blocant ou pas
	 */
	public function getBlockingMode(){
		return $this->default_blocking_mode;
	}
}

/**
        @brief Classe de base pour toutes les classes permettant de générer un objet WSSocket
        @details Cette classe encapsule les options SSLs, les protocols supportés et le FragmentSize par défaut.
*/
abstract class WSInitiator extends Threaded{
	/** @brief Vaut null si on utilise pas SSL, sinon contient un array avec les paramètres SSLs.*/
	protected $ssl_options=null;

	/** @brief Contient un array avec les protocols WebSockets supportés.*/
	protected $supported_protocols=null;

	/** @brief Contient le FragmentSize par defaut de toutes les WSSocket qui seront générés.*/
	private $default_ws_fragment_size=1300;

	/** @brief Initalise l'initiateur/générateur avec ses protocols supportés et les options SSLs si on souhaite utiliser SSL.
		@param array $supported_protocols Liste de protocols supportés.
		@param array $ssl_options Array\<string,string\> comportant les options SSLs sous la forme option_name => option_value.
	*/
	public function __construct($supported_protocols,$ssl_options=null){
		$this->supported_protocols=$supported_protocols;
		if( is_array ($ssl_options) ) $this->ssl_options=$ssl_options;
	}

	/**
		@brief Permet de savoir si les sockets générées seront SSL ou pas (elles seront alors de type TCP).
		@retval boolean TRUE si SSL, FALSE si non chiffré.
	*/
	public function isSSLReady(){
		return ( $this->ssl_options != null );
	}

	/**
		@brief Si WSInitiator::isSSLReady() vaut TRUE, congigure SSL sur la WSSocket passée en paramète.
		@param WSSocket $socket Une socket sur laquelle on souhaite configurer SSL.
	*/
	protected function initSSL($socket){
		if($this->isSSLReady()){
			foreach($this->ssl_options as $key => $value )  $socket->setContextOption($key,$value);
		}
	}

	/**
		@brief Permet de récuperer la liste des protocols supportés.
		@retval array Liste des protocols supportés.
	*/
	public function getProtocolsList(){
		return (array)$this->supported_protocols;
	}

	/**
		@brief Permet de checker si un protocol est connu de l'initiator
		@param string $name Nom du protocol dont ont souhaite vérifier la présence.
		@retval boolean TRUE si le protocol est connu, FALSE dans le cas contraire.
	*/
	public function isProtocolExist($name){
		return in_array($name,$this->getProtocolsList());
	}

	/**
		@brief Permet de choisir un protocol en comparant 2 lists.
		@details Cette méthode compare la liste $liste avec la liste WSInitiator::$supported_protocols afin de déterminer quel protocol sera choisi. Si plusieurs protocols correspondent, alors on utilise l'ordre de la liste WSInitiator::$supported_protocols pour déterminer le protocol qui sera retournée. Le premier de la liste WSInitiator::$supported_protocols étant le plus prioritaire et le dernier le moins prioritaire.
		@param array $list Liste des protocols qu'on souhaite pouvoir utiliser.
		@retval null|string Si aucun protocol correspond, alors null est retourné. Sinon, ce sera le protocol le plus prioritaire de la liste WSInitiator::$supported_protocols qui est contenu dans la liste $list.
	*/
	protected function acceptProtocols($list){
		$supported=$this->getProtocolsList();
		$accepted_protocol = null;
		for( $i=0; $i < count($supported) && $accepted_protocol == null; $i++){
			if( in_array( $supported[$i], $list ) ) $accepted_protocol=$supported[$i];
		}
		return $accepted_protocol;
	}

	/**
		@brief Permet de configurer le FragmentSize par défaut qu'auront toutes les WSSocket générées.
		@param int $size La taille que doit avoir le FragmentSize par défaut.
	*/
	public function setDefaultFagmentSize($size){
		$this->default_ws_fragment_size=$size;
	}

	/**
		@brief Récupére le FragmentSize par défaut attribué à toutes les WSSocket générées.
		@retval int La taille des FragmentSize des WSSocket générées.
	*/
	public function getDefaultFragmentSize(){
		return $this->default_ws_fragment_size;
	}

	/**
		@brief Permet de récupérer le sha1 d'une clé encodé en base64.
		@param string $key La clé qu'on souhaite encoder.
		@retval string L'empreinte sha1 de cla clé $key encodé en base64.
	*/
	protected static function encode_key($key){
		return base64_encode(sha1($key, true));
	}

	/**
		@brief Retourne le string permettant de signifier qu'$encoded_key est accepté.
		@param string $encoded_key Le sha1 d'une clé encodé en base64.
		@retval string Un sha1 encodé en base64 qui permet de signifier que la clé est acceptée.
	*/
	protected static function accept_key($encoded_key){
		$acceptKey = $encoded_key.'258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
        $acceptKey = self::encode_key($acceptKey);
		return $acceptKey;
	}
}

/**
        @brief Classe Permettant d'initier des WSSocket sur une application serveur.
        @details Cette classe manipule un StreamSocket binder sur une addresse et un port. Lorsque un client se connecte et qu'un WSServerInitiator::WSServerInitiator() est effectué, cette classe réalise le handshake WebSocket et génére une WSSocket qui permettra la communication avec le client.
*/
class WSServerInitiator extends WSInitiator{
	/** @brief Uniquement les clients possédant une des clés de cette liste seront acceptés */
	private $server_keylist;

	/** @brief La StreamSocket permettant de se binder sur un port */
	private $master_socket;

	/**
		@brief Permet de savoir si la StreamSocket est bindé sur le port
		@retval TRUE si la StreamSocket est créée et est bindé. FALSE dans le cas contraire.
	*/
	public function isBinded(){
		return (is_a($this->master_socket,"SocketStream") && !$this->master_socket->isConnected());
	}

	/**
		@brief Initialise Le WSServerInitiator avec la liste de clés accetpées, la liste des protocols acceptés et, pour SSL, la liste des options SSLs.
		@param array $accepted_keys Liste des clés qui seront accepté pour le handshake.
		@param array $accepted_protocols Liste des protocols qui seront acceptés dans le handshake. L'ordre de la liste correspond à l'ordre de préférence.
		@param array $ssl_options Tableau de type array<string,string> correspondant à $option_name => $option_value permettant d'activer et configurer le SSL.
	*/
	public function __construct($accepted_keys,$accepted_protocols,$ssl_options=null){
		$encoded_key_list=array();
		foreach($accepted_keys as $value){ array_push($encoded_key_list,self::encode_key($value)); }
		$this->server_keylist=(array)$encoded_key_list;
		$this->master_socket=null;
		parent::__construct($accepted_protocols,$ssl_options);
	}	

	/**
		@brief Crée la SocketStream, la configure et la bind sur le port.
		@param string $addr L'adresse IP d'écoute.
		@param int $port Le port d'écoute.
	*/
	public function bind($addr,$port){
		$this->master_socket = new SocketStream($this->isSSLReady());
		if($this->isSSLReady()) $this->initSSL($this->master_socket);
		$this->master_socket->bind($addr,$port);
	}

	/** 
		@brief Se mettre en attente d'une connection d'un client
		@details Permet de se mettre en attente de la connection d'un client et ensuite de réaliser le handshake. Si cela réussi, la valeur de retour sera la WSSocket permet de communiquer avec le client.
		@param int $timeout Le timeout en secondes pour l'acceptation de la connection. Voir StreamSocket::accept().
		@retval SocketStream La SocketStream permet de communiquer avec le peer.
		@exception WebSocketSSLException Si on souhaite une Socket SSL et que l'accept à donné une StreamSocket non SSL.
		@exception BadHeaderException Si le header HTTP n'est pas un Header compatible WebSocket.
		@exception BadVersionException Si la version WebSocket est incorrecte.
		@exception BadKeysException Si la clé envoyé par le client n'est pas acceptée.
		@exception BadProtocolsException Si aucun protocol envoyé par le client n'est connu.
	*/
	public function waitConnection($timeout=null,$notThreaded = false){
		$client_socket=$this->master_socket->accept($timeout,$notThreaded);

		if( $this->isSSLReady() && !$client_socket->isEncrypted() )
			throw new WebSocketSSLException("The SSL socket isn't encrypted");

		$headers = $client_socket->receive(2048);
		if(preg_match("/Sec-WebSocket-Version: (.*)\r\n/i", $headers, $match))  $version = $match[1];
			else throw new BadHeaderException("The client doesn't support WebSocket");
		if($version != 13) throw new BadVersionException("WebSocket version 13 required (the client supports version {$version})");
		
		// Extract header variables
		if(preg_match("/GET (.*) HTTP/", $headers, $match))   			$root = $match[1];
		if(preg_match("/Host: (.*)\r\n/i", $headers, $match))  			$host = $match[1];
		if(preg_match("/Origin: (.*)\r\n/i", $headers, $match))			$origin = $match[1];
		if(preg_match("/Sec-WebSocket-Key: (.*)\r\n/i", $headers, $match))	$received_key = $match[1];
		if(preg_match_all("/Sec-WebSocket-Protocol: (.*)\r\n/iU", $headers, $match, PREG_SET_ORDER)){
			$client_protocols=array();

			if( count($match) == 1 ) $client_protocols=explode(",",$match[0][1]);
			else foreach ($match as $protocol ) { array_push($client_protocols,$protocol); }

			for($i=0;$i<count($client_protocols);$i++){
				$client_protocols[$i]=trim($client_protocols[$i]);
			}
		}
		else throw new BadHeaderException("Sec-WebSocket-Protocol not found in header");
				
		//Acceptation de la clé
		if(!in_array( $received_key , $this->server_keylist ) ) throw new BadKeysException("Received key unknown");
		$acceptKey = self::accept_key($received_key);

		//Acceptation du protocol
		$accepted_protocol=$this->acceptProtocols($client_protocols);
		if( $accepted_protocol == null ) throw new BadProtocolsException("Unknown websocket-protocol");

		//Response header
		$upgrade = "HTTP/1.1 101 Switching Protocols\r\n".
                   "Upgrade: websocket\r\n".
                   "Connection: Upgrade\r\n".
                   "Sec-WebSocket-Accept: $acceptKey\r\n".
			       "Sec-WebSocket-Protocol: $accepted_protocol".
                   "\r\n\r\n";
		$client_socket->send($upgrade);
		if($notThreaded === false)
		{
			return new WSSocket($accepted_protocol,$client_socket,$this->getDefaultFragmentSize());
		}
		else 
		{
			return new WSSocketNotThreaded($accepted_protocol,$client_socket,$this->getDefaultFragmentSize());
		}
	}

	/**
		@brief Ferme la StreamSocket si celle-ci existe est n'est pas encore fermée.
	*/
	public function close(){
		if($this->isBinded()) $this->master_socket->close();
	}

}

/**
        @brief Classe Permettant d'initier des WSSocket sur une application cliente.
        @details Cette classe permet de réaliser un WSSocket handshake sur un serveur WebSocket et ensuite de retourner un WSSocket pour communiquer avec ce serveur.
*/
class WSClientInitiator extends WSInitiator{
	/** @brief La clé du client */
	private $client_key;

	/**
		@brief Initialise Le WSClientInitiator avec sa clé, la liste des protocols acceptés et, pour SSL, la liste des options SSLs.
		@param string $client_key Clé qui doit être connue par le serveur.
		@param array $accepted_protocols Liste des protocols qui seront acceptés dans le handshake. C'est le serveur qui choisira le protocole qui sera utilisé.
		@param array $ssl_options Tableau de type array<string,string> correspondant à $option_name => $option_value permettant d'activer et configurer le SSL.
	*/
	public function __construct($client_key,$accepted_protocols,$ssl_options=null){
		$this->client_key=self::encode_key($client_key);
		parent::__construct($accepted_protocols,$ssl_options);
	}	

	/** 
		@brief Permet d'initier une connection avec un serveur.
		@details Permet d'initier une connection avec un serveur. et ensuite de réaliser le handshake. Si cela réussi, la valeur de retour sera la WSSocket permet de communiquer avec le serveur.
		@param string $url L'url du serveur, sous la forme ws://\<host\>:\<port\>.	
		@param int $timeout Le timeout en secondes pourla connection. Voir StreamSocket::connect().
		@retval SocketStream La SocketStream permet de communiquer avec le peer.
		@exception WebSocketSSLException Si on souhaite une Socket SSL et que l'accept à donné une StreamSocket non SSL.
		@exception BadHeaderException Si le header HTTP n'est pas un Header compatible WebSocket.
		@exception BadHttpResponseException Si la réponse du serveur n'est pas une réponse WebSocket correcte.
		@exception BadKeysException Si la clé envoyé par le client n'est pas acceptée.
		@callgraph
	*/
	public function initiateConnection($url,$timeout=null){
		//URL parsing see : https://github.com/Textalk/websocket-php/blob/master/lib/Client.php#L45
		$url_parts = parse_url($url);
    	$scheme    = $url_parts['scheme'];
    	$host      = $url_parts['host'];
//    	$user      = isset($url_parts['user']) ? $url_parts['user'] : '';
//    	$pass      = isset($url_parts['pass']) ? $url_parts['pass'] : '';
    	$port      = isset($url_parts['port']) ? $url_parts['port'] : ($scheme === 'wss' ? 443 : 80);
    	$path      = isset($url_parts['path']) ? $url_parts['path'] : '/';
    	$query     = isset($url_parts['query'])    ? $url_parts['query'] : '';
    	$fragment  = isset($url_parts['fragment']) ? $url_parts['fragment'] : '';

		$path_with_query = $path;
    	if (!empty($query))    $path_with_query .= '?' . $query;
    	if (!empty($fragment)) $path_with_query .= '#' . $fragment;

		//Génération du header
		$supported_protocols=implode(", ",$this->getProtocolsList());

		$headers = array(
			'Host'		=> $host . ":" . $port,
			'User-Agent'    => 'websocket-client-php',
			'Connection'	=> 'Upgrade',
			'Upgrade'	=> 'websocket',	
			'Sec-WebSocket-Key'     => $this->client_key,
			'Sec-WebSocket-Version' => '13',
			'Sec-WebSocket-Protocol'  => $supported_protocols,
		);
		$header = implode("\r\n",array_map( function($key, $value) { return "$key: $value"; }, array_keys($headers), $headers));
		$header = "GET " . $path_with_query . " HTTP/1.1\r\n$header\r\n\r\n";

		//Ouverture de la connection
	//	Threaded::extend(SocketStream::class);
		$socket=new SocketStream($this->isSSLReady());
		if($this->isSSLReady()) $this->initSSL($socket);
		$socket->connect($host,$port,$timeout);	

		if( $this->isSSLReady() && !$socket->isEncrypted() )
			throw new WebSocketSSLException("The SSL socket isn't encrypted");

		//Handshake
		$socket->send($header);
		$headers=$socket->receive(2048);

		// Extract header variables
		if(preg_match("#HTTP/1\.1\s+(\d+)\s+(.*)\r\n#", $headers, $match))  {
			$http_response_code = $match[1];
			$http_response_message=$match[2];
		}	
		
		if( empty( $http_response_code ) ) throw new BadHeaderException("Unable to get http Response");
		if( $http_response_code != 101 ) throw new BadHttpResponseException($http_response_code,$http_response_message,101,"Switching Protocols");
        if(!preg_match("/Upgrade: websocket\r\n/i", $headers, $match))  
			throw new BadHTTPResponseException("Erreur in HttpResponse Upgrade field");
        if(!preg_match("/Connection: Upgrade\r\n/i", $headers, $match))                
			throw new BadHTTPResponseException("Erreur in HttpResponse Connection field");
        if(preg_match("/Sec-WebSocket-Accept: (.*)\r\n/i", $headers, $match))	$response_key = $match[1];
		if(preg_match("/sec-WebSocket-Protocol: (.*)\r\n/i",$headers, $match)) 	$response_protocol = $match[1];

		if( self::accept_key($this->client_key) != $response_key )  throw new BadKeysException("Erreur with the accept key");
		
		return new WSSocket($response_protocol,$socket,$this->getDefaultFragmentSize());
	}
}
?>
